#!/bin/bash

# http://www.opennet.ru/base/sec/port_knocking.txt.html
#   :
#  -       ""     
#     (port0:port+4095)
#  -            
#    (  )
#  -       n  (  100)
#  -     (   n 
#        )
#  -     ssh      
#  -      ESTABLISHED   ssh 

#    firewall
#    syslog 

#   :
#   -       
#          (  100 )
#   -   DoS


# base port of knocking (range from port0 to port0+4095 must be free)
port0=11981
# password
pass="password"
# unique string of knocking logs
id_string="PORT_KNOCKING"
# knocking log file
log_file="/var/log/kern.log"
# ip of our interface
our_ip="85.118.229.114"
# iptables binary
IPTABLES="/sbin/iptables"

# -------------------------------

# allowing only one connect from ip in this time period (seconds)
# and also in other words max period of client and server clock desynchronisation
time_period=200
# max time (in seconds) between two knocks in knocking sequence    
delta=2
# time period (in seconds) when door is open 
sleep_time=10

# don't touch
time_flag=0
used_time=0
cnt=0

# iptables initialisation for knocking listening
port1=$(($port0+4095))
# delete the LOG rule to prevent doublicate
$IPTABLES -D INPUT -s 0.0.0.0/0 -d $our_ip -p tcp --syn --dport $port0:$port1 -j LOG --log-level notice --log-prefix "$id_string "
# default policy is DROP
$IPTABLES -I INPUT 1 -s 0.0.0.0/0 -d $our_ip -p tcp --syn --dport $port0:$port1 -j LOG --log-level notice --log-prefix "$id_string "
# if policy is ACCEPT 
#$IPTABLES -A INPUT -p tcp --dport $port0:$port1 -j DROP

# allow only established ssh connection
#$IPTABLES -A INPUT -s 0.0.0.0/0 -d $our_ip -m state --state ESTABLISHED -p tcp --dport 22 -j ACCEPT
$IPTABLES -I inet 1 -s 0.0.0.0/0 -d $our_ip -m state --state ESTABLISHED -p tcp --dport 22 -j ACCEPT
# if policy is ACCEPT
$IPTABLES -D inet -p tcp --sport 1024:65535 --dport 22 -m state --state NEW -j ACCEPT

tail -n1 --follow=name --retry $log_file | 
{
    # read no using line
    read
    
    # main cycle of reading log lines 
    while [ 1 == 1 ]
    do
        read str

        # get time in seconds since `00:00:00 1970-01-01 UTC'
        time=`date +%s`

        # check is it our log line 
        ok=`echo $str | grep $id_string`
        if [ -z "$ok" ]; then
            # to next iteration of main cycle
            continue
        fi 

        # extract source ip and destination port from log line
        for fld in $str
        do
            case "${fld:0:4}" in
                "SRC=")
                        sip=${fld:4}
                        ;;
                "DPT=")
                        dport=${fld:4}
            esac
        done    

        
        # calculate secure combination of ports and time up to which this combinaton is valid
        if [ $time -ge $used_time ]; then
            time_stamp=$(($time/$time_period))    
            sum=`echo $pass$time_stamp | md5sum`
            i=0
            sec_ports=""
            while [ $i -lt 16 ]
            do
                j=${sum:$i*2:2}
                port=$(($port0+0x$j*16+$i))
                sec_ports="$sec_ports $port"
                i=$((i+1))
            done

            remainder=$(($time%$time_period))
            used_time=$(($time-$remainder+$time_period))
            used_ips=""

        fi

        # time period from last successsful processing
        dtime_flag=$(($time-$time_flag))
        
        if [ $dtime_flag -gt $delta -o $cnt -eq 0 ]; then 

            # check if our ip already processed in current time period
            ok=`echo $used_ips | grep $sip`
            if [ "$ok" ]; then
                # to next iteration of main cycle
                continue
            fi
            
            # begin processing for this ip
            cur_ip=$sip
            ports=""
            cnt=0
        
        else
            # not allowed simultaneously process more then one ip       
            if [ $sip != $cur_ip ]; then
                # to next iteration of main cycle
                continue
            fi
        fi

        # time label of successful processing
        time_flag=$time
        # list of processing ports
        ports="$dport $ports"
        # port counter
        cnt=$((cnt+1))

        # it's time to check port sequence
        if [ $cnt -eq 16 ]; then
            cnt=0

            # check if incoming knocking correct
            for port in $sec_ports
            do
                ok=`echo $ports | grep $port`
                if [ -z "$ok" ]; then
                    continue
                fi 
            done
            
            # open our door for some time
            if [ "$ok" ]; then 
                used_ips="$used_ips $cur_ip"

                # turn on incoming ssh connects

                # if policy is ACCEPT           
                #$IPTABLES -D inet -p tcp --dport 22 -j DROP

                $IPTABLES -I inet 2 -s $cur_ip -d $our_ip -p tcp --syn --dport 22 -j ACCEPT

                # if policy is ACCEPT           
                #$IPTABLES -A inet -p tcp --dport 22 -j DROP
                sleep $sleep_time
                # turn off incoming ssh connects
                $IPTABLES -D inet -s $cur_ip -d $our_ip -p tcp --syn --dport 22 -j ACCEPT
            fi
        fi

    done
}

